--- Implementation of Arrays with Braun Trees. Conceptually, Braun trees
--- are always infinite. Consequently, there is no test on emptiness.
--- @authors {bbr, fhu}@informatik.uni-kiel.de

module Array 
  (Array,

   emptyErrorArray, emptyDefaultArray,
   listToDefaultArray,listToErrorArray,

   (//), update, applyAt,

   (!))

  where 


import Integer

infixl 9  !, //

data Array b = Entry b (Array b) (Array b)


--- Creates an empty array which generates errors for non-initialized
--- indexes.
emptyErrorArray :: Array b
emptyErrorArray = defaultArray errorArray 0 1

errorArray :: Int -> _
errorArray idx =  error ("Array index "++show idx++" not initialized")

--- Creates an empty array, call given function for non-initialized
--- indexes.
--- @param default - function to call for each non-initialized index
emptyDefaultArray :: (Int -> b) -> Array b
emptyDefaultArray default = defaultArray default 0 1

defaultArray :: (Int -> b) -> Int -> Int -> Array b
defaultArray default b o =
  Entry (default (b+o-1))
	(defaultArray default (2*b) o)
        (defaultArray default (2*b) (o+b))

--- Inserts a list of entries into an array.
--- @param array - array to modify
--- @param modifications - list of new (indexes,entries)
--- If an index in the list was already initialized, the old value
--- will be overwritten. Likewise the last entry with a given index
--- will be contained in the result array.
(//) :: Array b -> [(Int,b)] -> Array b
array // modifications = foldr (\ (n,v) a -> update a n v)
                               array
                               modifications

--- Inserts a new entry into an array.
--- @param array - array to modify
--- @param idx - index of update
--- @param val - value to update at index idx
--- Entries already initialized will be overwritten.
update :: Array b -> Int -> b -> Array b
update a i v = applyAt a i (const v)

--- Applies a function to an element.
--- @param array - array to modify
--- @param idx - index of update
--- @param fun - function to apply on element at index idx
applyAt :: Array b -> Int -> (b->b) -> Array b
applyAt (Entry v al ar) n f
  | n==0      = Entry (f v) al ar
  | odd n     = Entry v (applyAt al (n `div` 2) f) ar
  | otherwise = Entry v al (applyAt ar (n `div` 2 - 1) f)


--- Yields the value at a given position.
--- @param a - array to look up in
--- @param n - index, where to look 
(!) :: Array b -> Int -> b
(Entry v al ar) !n 
  | n==0      = v
  | odd n     = al!(n `div` 2)
  | otherwise = ar!(n `div` 2 - 1)


split :: [a] -> ([a],[a])
split [] = ([],[])
split [x] = ([x],[])
split (x:y:xys) = let (xs,ys) = split xys in
                    (x:xs,y:ys)

--- Creates a default array from a list of entries.
--- @param def - default funtion for non-initialized indexes
--- @param xs - list of entries
listToDefaultArray ::  (Int -> b) -> [b] -> Array b
listToDefaultArray def = 
  listToArray (defaultArray def) 1 0 

--- Creates an error array from a list of entries.
--- @param xs - list of entries
listToErrorArray :: [b] -> Array b
listToErrorArray = listToArray errorArray 1 0 


listToArray :: (Int -> Int -> Array b) -> Int -> Int -> [b] -> Array b
listToArray def b o [] = def b o
listToArray def b o (x:xs) = let (ys,zs) = split xs in
                         Entry x (listToArray def (2*b) o ys)
                                 (listToArray def (2*b) (o+b) zs)

